from __future__ import annotations

import json
import logging
from enum import StrEnum
from html import unescape
from pathlib import Path
from typing import Self
from urllib.parse import parse_qs, urlparse

import aiofiles
import m3u8
from httpx import AsyncClient
from pydantic import BaseModel, Field, HttpUrl, computed_field

from pyokru.errors import (
    CopyrightsRestrictedError,
    NotReadyError,
    UnknownError,
    VideoNotFoundError,
    VideoProcessingError,
)
from pyokru.utils import gather_with_concurrency, parse_json, search_regex
from pyokru.utils.download import download_m3u8_format, download_mp4_file
from pyokru.utils.formats import get_mpd_qualities_number, get_video_dimensions

CHUNK_SIZE = 15 * (1024 * 1024)  # 15MB


def slice_per(source: int, step: int) -> list[tuple[int, int]]:
    def maxer(file_size: int, value: int) -> int:
        return value if value <= file_size else file_size

    return [(i, maxer(source, i + step)) for i in range(0, source, step)]


class SourceType(StrEnum):
    M3U8 = "m3u8"
    MPD = "mpd"
    MP4 = "mp4"


class Quality(BaseModel):
    name: str
    source_type: SourceType
    resolution: tuple[int, int] | None = None
    local_resolution: tuple[int, int] | None = None
    url: str = Field(description="Quality's url")
    video: Video = Field(exclude=True, repr=False)

    async def download(self: Self, path: Path) -> None:
        match self.source_type:
            case SourceType.MP4:
                await download_mp4_file(
                    file_path=path,
                    url=self.url,
                    video_id=self.video.id,
                )
            case SourceType.M3U8:
                await download_m3u8_format(
                    file_path=path,
                    m3u8_link=self.url,
                    video_id=self.video.id,
                )
        self.local_resolution = await get_video_dimensions(path)


class Video(BaseModel):
    id: str
    delete_id: int | None = None
    title: str
    file_path: Path
    m3u8_url: HttpUrl | None = None
    mpd_url: HttpUrl | None = None
    duration: int | None = None
    copyrights_restricted: bool = False
    qualities: list[Quality] = []
    expected_qualities_number: int = None

    @computed_field
    @property
    def url(self: Video) -> HttpUrl:
        return HttpUrl(f"https://ok.ru/video/{self.id}")

    async def fetch_info(
        self: Video, session_headers: dict[str, str], only_one_type: bool = False
    ) -> bool:
        session_headers["Accept-Language"] = "en-US,en;q=0.9"
        async with AsyncClient(
            headers=session_headers, http2=True, timeout=15
        ) as client:
            video_id = str(self.id)
            html_player = await client.get(self.url.unicode_string())

            detected_player = search_regex(
                rf'data-options=(?P<quote>["\'])(?P<player>{{.+?{video_id}.+?}})(?P=quote)',
                html_player.text,
                group="player",
            )
            error = search_regex(
                r'[^>]+class="vp_video_stub_txt"[^>]*>([^<]+)<', html_player.text
            )
            match error:
                case "Video has not been found":
                    raise VideoNotFoundError
                case "An error occurred while processing the video.":
                    raise VideoProcessingError
                case "The video is not ready, please wait.":
                    raise NotReadyError
                case (
                    "Video has been blocked due to author's rights infingement"
                    | "Видео заблокировано из-за нарушений авторских прав"
                ):
                    raise CopyrightsRestrictedError
                case None:
                    pass
                case _:
                    raise UnknownError
            self.qualities = []
            player = parse_json(unescape(detected_player)) if detected_player else None
            if not player:
                return False
            metadata: dict = json.loads(player["flashvars"].get("metadata"))
            if metadata["movie"]["status"] == "COPYRIGHTS_RESTRICTED":
                self.copyrights_restricted = True
            self.duration: int = metadata["movie"]["duration"]
            mpd_url = metadata.get("ondemandDash")
            m3u8_url = metadata.get("ondemandHls") or metadata.get("hlsManifestUrl")
            mp4_urls = metadata.get("videos")
            for quality in mp4_urls:
                self.qualities.append(
                    Quality(
                        name=quality["name"],
                        url=quality["url"],
                        source_type=SourceType.MP4,
                        video=self,
                    )
                )
            if not mpd_url:
                return True
            if not self.expected_qualities_number:
                self.mpd_url = HttpUrl(mpd_url)
                self.expected_qualities_number = await get_mpd_qualities_number(
                    client, mpd_url, self.id
                )
            if only_one_type:
                self.qualities = []
            self.m3u8_url = HttpUrl(m3u8_url)
            self.qualities += await _extract_qualities(
                client, self.m3u8_url.unicode_string(), self
            )
            m3u8_qualities_no = (
                len(self.qualities)
                if only_one_type
                else (len(self.qualities) - len(mp4_urls))
            )
            return not m3u8_qualities_no < self.expected_qualities_number

    def delete_file(self: Video) -> None:
        self.file_path.unlink(missing_ok=True)


async def _extract_qualities(
    client: AsyncClient, m3u8_url: str, video: Video
) -> list[Quality]:
    base_url = m3u8_url.rsplit("/", 1)[0]
    m3u8_content = (await client.get(m3u8_url)).text
    info = m3u8.loads(m3u8_content)
    qualities: list[Quality] = []
    for p in info.data["playlists"]:
        stream_info = p["stream_info"]
        qualities.append(
            Quality(
                name=stream_info["quality"],
                resolution=stream_info["resolution"].split("x"),
                source_type=SourceType.M3U8,
                url=f"{base_url}{'' if p['uri'].startswith('/') else '/'}{p['uri']}",
                video=video,
            )
        )
    return list(sorted(qualities, key=lambda x: x.resolution))


async def _delete_video(session: str, headers: dict[str, str], vid: int) -> None:
    async with AsyncClient(
        base_url="https://api.ok.ru", headers=headers, http2=True, timeout=15
    ) as client:
        params = {
            "__screen": "feed_main",
            "application_key": "CBAFJIICABABABABA",
            "session_key": session,
            "vid": vid,
        }
        await client.post("/api/video/delete", params=params)


async def _upload_video(
    session: str, headers: dict[str, str], name: str, file: Path
) -> Video:
    async with AsyncClient(
        base_url="https://api.ok.ru", headers=headers, http2=True, timeout=15
    ) as client:
        file_size = file.stat().st_size
        params = {
            "__screen": "feed_main",
            "application_key": "CBAFJIICABABABABA",
            "session_key": session,
            "file_name": name,
            "file_size": file_size,
        }
        data = (await client.get("/api/video/getUploadUrl", params=params)).json()
        video_id: int = data["video_id"]
        params = {
            "__screen": "video_feeds,video_showcase",
            "application_key": "CBAFJIICABABABABA",
            "session_key": session,
            "description": "",
            "privacy": "FRIENDS",
            "vid": video_id,
        }
        await client.get("/api/video/update", params=params)
        upload_url: str = data["upload_url"]
        final_video_id = parse_qs(urlparse(upload_url).query)["id"][0]
        upload_chunk_tasks = []

        for content_range in slice_per(file_size, CHUNK_SIZE):
            start_at, end_at = content_range

            upload_chunk_tasks.append(
                _upload_chunk(client, upload_url, start_at, end_at, file)
            )
        await gather_with_concurrency(4, upload_chunk_tasks)
        return Video(id=final_video_id, delete_id=video_id, title=name, file_path=file)


async def read_chunk(file: Path, start_index: int, end_index: int) -> bytes:
    async with aiofiles.open(file, "rb") as fin:
        await fin.seek(start_index)
        return await fin.read(end_index - start_index)


async def _upload_chunk(  # noqa: PLR0913
    client: AsyncClient,
    upload_url: str,
    index: int,
    offset: int,
    file: Path,
) -> None:
    chunk = await read_chunk(file, index, offset)
    for _ in range(3):
        try:
            res = await client.post(
                upload_url,
                data=chunk,
                headers={
                    "Content-Range": (
                        f"bytes {index}-{offset - 1}/{file.stat().st_size}"
                    ),
                    "X-Uploading-Mode": "parallel",
                    "Content-Disposition": f'attachment; filename="{file.name}"',
                },
                timeout=30,
            )
            if res.status_code in [201, 200]:
                break
        except Exception:
            logging.exception("OKRU")
