from __future__ import annotations

import asyncio
import json
import re
from collections.abc import Awaitable, Callable, Iterable
from typing import Any


class LenientJSONDecoder(json.JSONDecoder):
    # TODO: Write tests
    def __init__(
        self: LenientJSONDecoder,
        *args,  # noqa: ANN002
        transform_source=None,  # noqa: ANN001
        ignore_extra: bool = False,
        close_objects: int = 0,
        **kwargs,  # noqa: ANN003
    ) -> None:
        self.transform_source, self.ignore_extra = (
            transform_source,
            ignore_extra,
        )
        self._close_attempts = 2 * close_objects
        super().__init__(*args, **kwargs)

    @staticmethod
    def _close_object(err):  # noqa: ANN205, ANN001
        doc = err.doc[: err.pos]
        # We need to add comma first to get the correct error message
        if err.msg.startswith("Expecting ','"):
            return doc + ","
        elif not doc.endswith(","):  # noqa: RET505
            return None

        if err.msg.startswith("Expecting property name"):
            return doc[:-1] + "}"
        elif err.msg.startswith("Expecting value"):  # noqa: RET505
            return doc[:-1] + "]"
        return None

    def decode(self: LenientJSONDecoder, s):  # noqa: ANN001, ANN201
        if self.transform_source:
            s = self.transform_source(s)
        for attempt in range(self._close_attempts + 1):
            try:
                if self.ignore_extra:
                    return self.raw_decode(s.lstrip())[0]
                return super().decode(s)
            except json.JSONDecodeError as e:  # noqa: PERF203
                if e.pos is None:
                    raise
                elif attempt < self._close_attempts:  # noqa: RET506
                    s = self._close_object(e)
                    if s is not None:
                        continue
                raise type(e)(  # noqa: B904, TRY200, TRY003
                    f"{e.msg} in {s[e.pos-10:e.pos+10]!r}", s, e.pos
                )
        raise AssertionError("Too many attempts to decode JSON")  # noqa: TRY003


def search_regex(
    pattern: str,
    string: str,
    flags: re._FlagsType = 0,
    group: None | tuple | list | str = None,
) -> str | tuple[str]:
    """
    Perform a regex search on the given string, using a single or a list of
    patterns returning the first matching group.
    In case of failure return a default value or raise a WARNING or a
    RegexNotFoundError, depending on fatal, specifying the field name.
    """
    if string is None:
        mobj = None
    elif isinstance(pattern, str | re.Pattern):
        mobj = re.search(pattern, string, flags)
    else:
        for p in pattern:
            mobj = re.search(p, string, flags)
            if mobj:
                break

    if mobj:
        if group is None:
            # return the first matching group
            return next(g for g in mobj.groups() if g is not None)
        elif isinstance(group, list | tuple):  # noqa: RET505
            return tuple(mobj.group(g) for g in group)
        else:
            return mobj.group(group)
    return None


def parse_json(json_string: str) -> dict:
    return json.loads(
        json_string,
        cls=LenientJSONDecoder,
        strict=False,
        transform_source=None,
    )


ASYNC_FUNC = Callable[[Any, Any], Awaitable[Any]]


async def gather_with_concurrency(
    number_of_parller_tasks: int, tasks: Iterable[ASYNC_FUNC[Any]]
) -> list[Any]:
    """
    run many functions in parallel

    Args:
        number_of_parller_tasks (int): Number of parallel tasks.

    Returns:
        List[Any]: The return of all given functions.
    """
    semaphore = asyncio.Semaphore(number_of_parller_tasks)

    async def sem_task(task: ASYNC_FUNC) -> Any:  # noqa: ANN401
        async with semaphore:
            return await task

    return await asyncio.gather(*(sem_task(task) for task in tasks))


def _str_cookies_to_dict(cookies_string: str) -> dict[str, str]:
    cookies_dict = {}
    for cookie in cookies_string.split(";"):
        key, value = cookie.split("=", maxsplit=1)
        cookies_dict[key.strip()] = value.strip()
    return cookies_dict
