import asyncio
import random
import shutil
import tempfile
from pathlib import Path

import aiofiles
import httpx
import m3u8
from pydantic import HttpUrl

from pyokru.utils import gather_with_concurrency

CHUNK_SIZE = 10 * 1024 * 1024  # 10MB

CONCURRENT_CONNECTIONS = 4


async def download_mp4_file(
    file_path: Path,
    url: HttpUrl,
    video_id: str,
) -> None:
    worker_folder = Path(tempfile.mkdtemp())

    headers = {
        "sec-ch-ua": (
            '"Chromium";v="122", "Not(A:Brand";v="24", "Microsoft Edge";v="122"'
        ),
        "sec-ch-ua-mobile": "?0",
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 Edg/123.0.0.0"
        ),
        "sec-ch-ua-platform": '"Windows"',
        "Accept": "*/*",
        "Origin": "https://ok.ru",
        "Sec-Fetch-Site": "cross-site",
        "Sec-Fetch-Mode": "cors",
        "Sec-Fetch-Dest": "video",
        "Referer": f"https://ok.ru/video/{video_id}",
    }
    async with httpx.AsyncClient(headers=headers) as client:
        response = await client.head(url)
        file_size = int(response.headers["Content-Length"])

        chunks = file_size // CHUNK_SIZE + (file_size % CHUNK_SIZE > 0)

        async def download_chunk(start: int, end: int, chunk_id: int) -> Path:
            chunk_response = await client.get(
                url, headers={"Range": f"bytes={start}-{end}"}
            )
            chunk_file_name = worker_folder / f"part_{chunk_id}"
            async with aiofiles.open(chunk_file_name, "wb") as f:
                await f.write(chunk_response.content)
            return chunk_file_name

        tasks = []
        for i in range(chunks):
            start = i * CHUNK_SIZE
            end = min(start + CHUNK_SIZE - 1, file_size - 1)
            tasks.append(download_chunk(start, end, i))

        chunk_files = await gather_with_concurrency(CONCURRENT_CONNECTIONS, tasks)
        async with aiofiles.open(file_path, "wb") as f:
            for chunk_file in chunk_files:
                async with aiofiles.open(chunk_file, "rb") as cf:
                    await f.write(await cf.read())
                chunk_file.unlink()
        shutil.rmtree(worker_folder, ignore_errors=True)


async def sleep_random() -> None:
    sleep_duration = random.uniform(0.06, 0.1)
    await asyncio.sleep(sleep_duration)


async def _download_segment(client: httpx.AsyncClient, url: str, folder: Path) -> int:
    filename = url.rsplit("/", 1)[-1]
    response = await client.get(url, timeout=30)

    if response.status_code == 500:  # noqa: PLR2004
        return response.status_code
    target_path = folder / filename
    async with aiofiles.open(target_path, "wb") as file:
        await file.write(response.content)
    return 200


async def _download_segment_handler(
    client: httpx.AsyncClient, url: str, folder: Path
) -> None:
    for _ in range(10):
        response_code = await _download_segment(client, url, folder)
        if response_code != 500:  # noqa: PLR2004
            return
        await sleep_random()


async def download_m3u8_format(
    file_path: Path,
    m3u8_link: str,
    video_id: str,
    number_of_parallel_download: int = 30,
) -> Path:
    async with httpx.AsyncClient(
        headers={
            "User-Agent": (
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 OPR/102.0.0.0"
            ),
        }
    ) as client:
        worker_folder = Path(tempfile.mkdtemp())

        m3u8_content = (await client.get(m3u8_link, timeout=30)).text
        m3u8_file_path = worker_folder / "base.m3u8"

        async with aiofiles.open(m3u8_file_path, "w") as ff:
            await ff.write(m3u8_content)
        m3u8_content = m3u8.loads(m3u8_content)
        m3u8_content.base_uri = m3u8_link.rsplit("/", 1)[0]

        download_client = httpx.AsyncClient(
            timeout=30,
            headers={
                "sec-ch-ua": (
                    '"Chromium";v="116", "Not)A;Brand";v="24", "Opera";v="102"'
                ),
                "sec-ch-ua-mobile": "?0",
                "User-Agent": (
                    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
                    "(KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36 OPR/102.0.0.0"
                ),
                "sec-ch-ua-platform": '"Windows"',
                "Accept": "*/*",
                "Origin": "https://ok.ru",
                "Sec-Fetch-Site": "cross-site",
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Dest": "empty",
                "Referer": f"https://ok.ru/video/{video_id}",
                "Accept-Language": "en-US,en;q=0.9",
            },
        )

        download_tasks = [
            _download_segment_handler(
                download_client, f"{segment.base_uri}/{segment.uri}", worker_folder
            )
            for segment in m3u8_content.segments
        ]
        await gather_with_concurrency(number_of_parallel_download, download_tasks)
        cmd = [
            "-v",
            "quiet",
            "-i",
            f"{m3u8_file_path}",
            "-c",
            "copy",
            "-y",
            f"{file_path}",
        ]
        proc = await asyncio.create_subprocess_exec("ffmpeg", *cmd)
        await proc.wait()
        shutil.rmtree(worker_folder, ignore_errors=True)
        return file_path
