import asyncio
import logging
import re
from pathlib import Path
from xml.etree import ElementTree

from httpx import AsyncClient
from pydantic import HttpUrl


def get_default_namespace(xml_content: str) -> str:
    try:
        default_ns_match = re.search(r'xmlns="([^"]+)"', xml_content)
        return default_ns_match.group(1) if default_ns_match else ""

    except ElementTree.ParseError:
        logging.exception("Error parsing XML:")
        return "def"


async def get_mpd_qualities_number(
    client: AsyncClient, mpd_link: HttpUrl, video_id: str
) -> int:
    headers = {
        "sec-ch-ua": (
            '"Microsoft Edge";v="123", "Not:A-Brand";v="8", "Chromium";v="123"'
        ),
        "sec-ch-ua-mobile": "?0",
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
            "(KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36 Edg/123.0.0.0"
        ),
        "sec-ch-ua-platform": '"Windows"',
        "Accept": "*/*",
        "Origin": "https://ok.ru",
        "Sec-Fetch-Site": "cross-site",
        "Sec-Fetch-Mode": "cors",
        "Sec-Fetch-Dest": "empty",
        "Referer": f"https://ok.ru/video/{video_id}",
        "Accept-Language": "ar",
    }
    response = await client.get(mpd_link, headers=headers, timeout=30)
    mpd_content = response.text
    namespaces = {"ns": get_default_namespace(mpd_content)}
    root = ElementTree.fromstring(mpd_content)
    video_qualities = []
    for adaptation_set in root.findall(".//ns:AdaptationSet", namespaces):
        mime_type = adaptation_set.get("mimeType")
        first_check = mime_type.startswith("video") if mime_type else False
        for representation in adaptation_set.findall(
            ".//ns:Representation", namespaces
        ):
            mime_type = representation.get("mimeType")
            second_check = mime_type.startswith("video") if mime_type else False
            if first_check or second_check:
                quality = representation.get("quality")
                video_qualities.append(quality)
    return len(video_qualities)


async def get_video_dimensions(video: Path) -> tuple[int, int] | None:
    cmd = [
        "-v",
        "error",
        "-select_streams",
        "v",
        "-show_entries",
        "stream=width,height",
        "-of",
        "csv=p=0:s=x",
        f"{video}",
    ]
    ps = await asyncio.subprocess.create_subprocess_exec(
        "ffprobe", *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await ps.communicate()
    stdout = stdout.decode().strip()
    if stdout:
        return [int(d) for d in stdout.split("x")]
    return None
